# Importing things to note:
# Scaling is not supported by BZ2 in either animations or transformation matrices. Use ctrl+A to apply scaling to the mesh in blender, this exporter will ignore scaling.
# If you want to export rigid body physics animations or any non-keyframe animations, remember to bake to keyframes in blender first.
# Do not flip or rotate your models for BZ2, this is done automatically with bpy_extras.io_utils.axis_conversion.
# No rescaling is done. 32.0 in blender is the size of one terrain square. (Like a guntower) 8.0 is the grid size, one terrain square is 4x4 of 8.0 with no bottom face.
# Vertex colors are not supported for meshes with bone envelopes in BZ2, but will be exported anyways.

# I've also made an importer: https://dl.dropboxusercontent.com/u/52091356/BZ2/Misc/Blender%20Exporter/io_import_bz2xsi.py

# This exporter supports the following:
# [.] xsi 0101txt 0032
# [.] SI_CoordinateSystem coord
# [ ] SI_Angle
# [ ] SI_Camera *
# [ ] SI_Ambience
# [ ] SI_Light *
# [ ] SI_Fog

# [.] Frame frm-*
# [.] Frame > FrameTransformMatrix
# [.] Frame > SI_FrameBasePoseMatrix
# [.] Frame > Mesh *
# [.] Frame > Mesh > MeshMaterialList
# [.] Frame > Mesh > MeshMaterialList > SI_Material
# [.] Frame > Mesh > MeshMaterialList > SI_Material > SI_Texture2D
# [.] Frame > Mesh > SI_MeshNormals
# [.] Frame > Mesh > SI_MeshVertexColors
# [.] Frame > Mesh > SI_MeshTextureCoords

# [.] AnimationSet
# [.] AnimationSet > Animation (anim-*)
# [.] AnimationSet > Animation > SI_AnimationKey

# [.] SI_EnvelopeList
# [.] SI_EnvelopeList > SI_Envelope

bl_info = {
	"name": "BZ2 XSI Model Format (.xsi)",
	"author": "The Vearidons (Stroket@hotmail.com)",
	"version": (1, 0, 7),
	"blender:": (2, 67, 0),
	"location": "File > Export > BZ2XSI (.xsi)",
	"description": "Export to a BZ2 compatible XSI format.",
	"warning": "Scaling is not supported in BZ2 and will not be exported. Ctrl+A to apply scaling.",
	"wiki_url": "http://www.meatspin.cc/",
	"tracker_url": "http://www.bz2maps.us/phpBB/viewtopic.php?f=50&t=4395/",
	"category": "Import-Export"}

import bpy
import re
from bpy.props import StringProperty, EnumProperty, BoolProperty
from mathutils import Matrix, Vector, Quaternion
from bpy_extras.io_utils import axis_conversion
from math import radians

class BZ2XSI_Settings:
	def __init__(self, context, FilePath, AddDummyRoot = 1, Indent = "", BoneMesh = False, ExportUV = True, ExportVertColors = True, ExportAnimations = True, ExportEnvelopes = True):
		self.context = context
		self.FilePath = FilePath
		self.AddDummyRoot = int(AddDummyRoot)
		self.Indent = Indent
		self.BoneMesh = BoneMesh
		
		self.ExportUV = ExportUV
		self.ExportVertColors = ExportVertColors
		self.ExportAnimations = ExportAnimations
		self.ExportEnvelopes = ExportEnvelopes
		
		self.IndentLvl = 0

# Axis Conversion
axisConv = axis_conversion(from_forward="Z", from_up="Y").to_4x4()

# Bones in an armature object aren't objects or sub-objects in blender.
# In BZ2 they are, this class is just an interface for simplicity to solve this.
# VType is the same as realObject.type, except in the case of bones, which it is "BONE",
# then realObject is a tuple containing the (Armature Object, Data Bone, and Pose Bone).
class Frm:
	def __init__(self, name, VType, realObject, parent):
		self.name = name
		self.VType = VType
		self.realObject = realObject
		
		# Default mesh: None.
		self.mesh = None
		
		# Default parent: parent.
		self.parent = parent
		
		# Get objects and sub-objects.
		self.children = []
		self.children += self.GetChildren(realObject)
		
		# Default animations: Nothing.
		self.Anim = [None, None]
		self.hasAnim = False
		
		# Default pose matrix: Nothing.
		self.poseMatrix = None
		
		# Default envelopes: Nothing.
		self.envelopes = {}
		
		if self.VType == "MESH" or VType == "EMPTY" or self.VType == "ARMATURE":
			# Matrix
			self.matrix = Matrix(self.realObject.matrix_local)
			
			# Mesh, if it's a mesh object.
			if self.VType == "MESH":
				self.mesh = self.realObject.to_mesh(bpy.context.scene, True, "PREVIEW")
			elif self.VType == "ARMATURE":
				# Create bones as Frm objects.
				for DataBone in self.realObject.data.bones:
					if not DataBone.parent:
						self.children += [Frm(DataBone.name, "BONE", (self.realObject, DataBone, self.realObject.pose.bones[DataBone.name]), self)]
			
			# Animation data.
			self.Anim = Animation(self.realObject, None, True)
			
			# Envelopes
			self.envelopes = self.GetObjEnvelopes()
		elif self.VType == "BONE":
			self.matrix = GetDataBoneMatrix(self.realObject[1])
			self.poseMatrix = GetPoseBoneMatrix(self.realObject[2])
			
			# Bone animation data
			self.Anim = Animation(self.realObject[0], self.realObject[2], True)
		if self.Anim.PosKeyFrameCount+self.Anim.RotKeyFrameCount:
			self.hasAnim = True
	def GetChildren(self, realObject):
		ChildList = []
		if self.VType == "BONE":
			for SubDataBone in realObject[1].children:
				ChildList += [Frm(SubDataBone.name, "BONE", (self.realObject[0], SubDataBone, self.realObject[0].pose.bones[SubDataBone.name]), self)]
		else:
			for RealObj in realObject.children:
				ChildList += [Frm(RealObj.name, RealObj.type, RealObj, self)]
		return ChildList
	def GetVertGroupVerts(self, Obj, Grp):
		VertList = []
		for Vert in Obj.data.vertices:
			for VGrp in Vert.groups:
				if VGrp.group == Grp.index:
					VertList += [Vert]
		return VertList
	def GetObjEnvelopes(self):
		Envelopes = {}
		for Grp in self.realObject.vertex_groups:
			EnvVertList = []
			EnvWeightList = []
			VertList = self.GetVertGroupVerts(self.realObject, Grp)
			for Vert in VertList:
				if Grp.weight(Vert.index):
					EnvVertList += [Vert.index]
					EnvWeightList += [Grp.weight(Vert.index)*100]
			Envelopes[Grp.name] = (EnvVertList, EnvWeightList)
		return Envelopes

# This class will contain simple access to all
# of the animation data, using the simple lists:
# PosKeyFrames[KeyFrame], RotKeyFrames[KeyFrame], Pos[Vector], Rot[Quaternion].
class Animation:
	def __init__(self, Obj, Bone = None, ForceEqLenAnim = True):
		self.Obj = Obj
		self.Bone = Bone
		self.ForceEqLenAnim = ForceEqLenAnim
		
		# Obtain all keyframes.
		KeyFrames = self.GetKeyframes()
		self.PosKeyFrames = KeyFrames[0]
		self.RotKeyFrames = KeyFrames[1]
		
		# Obtain animation data.
		Anims = self.GetAnimations(KeyFrames)
		self.Pos = Anims[0]
		self.Rot = Anims[1]
		
		if self.ForceEqLenAnim:
			if len(self.PosKeyFrames):
				if not self.PosKeyFrames[0] == bpy.context.scene.frame_start:
					self.PosKeyFrames = [bpy.context.scene.frame_start] + self.PosKeyFrames
					self.Pos = [self.Pos[0]] + self.Pos
				
				if not self.PosKeyFrames[-1] == bpy.context.scene.frame_end:
					self.PosKeyFrames += [bpy.context.scene.frame_end]
					self.Pos += [self.Pos[-1]]
			
			if len(self.RotKeyFrames):
				if not self.RotKeyFrames[0] == bpy.context.scene.frame_start:
					self.RotKeyFrames = [bpy.context.scene.frame_start] + self.RotKeyFrames
					self.Rot = [self.Rot[0]] + self.Rot
				
				if not self.RotKeyFrames[-1] == bpy.context.scene.frame_end:
					self.RotKeyFrames += [bpy.context.scene.frame_end]
					self.Rot += [self.Rot[-1]]
		
		self.PosKeyFrameCount = len(self.PosKeyFrames)
		self.RotKeyFrameCount = len(self.RotKeyFrames)
	def GetKeyframes(self):
		KeyFramesPos = []
		KeyFramesRot = []
		KFList = [KeyFramesPos, KeyFramesRot]
		
		if not self.Obj.animation_data or not self.Obj.animation_data.action:
			return KFList
		
		KF_Start = bpy.context.scene.frame_start
		KF_End = bpy.context.scene.frame_end
		if self.Bone:
			PathDataTypes = ("pose.bones[\"%s\"].location" % self.Bone.name, "pose.bones[\"%s\"].rotation_quaternion" % self.Bone.name)
		else:
			PathDataTypes = ("location", "rotation_euler")
		
		for FCurve in self.Obj.animation_data.action.fcurves:
			for PathDataIndex, PathDataType in enumerate(PathDataTypes):
				if FCurve.data_path == PathDataType:
					for KeyFrame in FCurve.keyframe_points:
						KFrm = int(KeyFrame.co[0])
						if KFrm >= KF_Start and KFrm <= KF_End and not KeyFrame.co[0] in KFList[PathDataIndex]:
							KFList[PathDataIndex] += [KFrm]
		
		KFList[0].sort()
		KFList[1].sort()
		return KFList
	def GetAnimations(self, KeyFrames):
		Position = []
		Rotation = []
		KeyFrameReset = bpy.context.scene.frame_current
		
		# Position
		for KeyFrm in KeyFrames[0]:
			bpy.context.scene.frame_set(KeyFrm)
			if self.Bone:
				Position += [GetPoseBoneMatrix(self.Bone).to_translation()]
			else:
				Position += [Matrix(self.Obj.matrix_local).to_translation()]
		
		# Rotation
		for KeyFrm in KeyFrames[1]:
			bpy.context.scene.frame_set(KeyFrm)
			# self.Bone is Real PoseBone or None.
			if self.Bone:
				Rotation += [GetPoseBoneMatrix(self.Bone).inverted().to_quaternion()]
			else:
				Rotation += [Matrix(self.Obj.matrix_local).to_quaternion()]
		
		bpy.context.scene.frame_set(KeyFrameReset)
		Animations = [Position, Rotation]
		return Animations

def ExportXSI(XSI):
	# All objects that exist as "Frm" class objects.
	AllObjects = CreateObjects()
	
	# Get initial, parentless frames.
	InitObjects = []
	for Obj in AllObjects:
		if Obj.parent == None:
			InitObjects += [Obj]
	
	if True:
		print("Exporting BZ2XSI: %s" % XSI.FilePath)
		XSI.File = open(XSI.FilePath, "w")
		
		# Write Header (XSI File info, Coordinate systems.)
		XSI.File.write("xsi 0101txt 0032\n\n")
		XSI.File.write("SI_CoordinateSystem coord {\n")
		XSI.File.write("%s1;\n" % XSI.Indent)
		XSI.File.write("%s0;\n" % XSI.Indent)
		XSI.File.write("%s1;\n" % XSI.Indent)
		XSI.File.write("%s0;\n" % XSI.Indent)
		XSI.File.write("%s2;\n" % XSI.Indent)
		XSI.File.write("%s5;\n" % XSI.Indent)
		XSI.File.write("}\n\n")
		
		# Add dummyroot if AddDummyRoot is 1.
		if XSI.AddDummyRoot == 1:
			XSI.File.write("Frame frm-hp_dummyroot {\n%sFrameTransformMatrix {\n%s1.000000,0.000000,0.000000,0.000000,\n%s0.000000,1.000000,0.000000,0.000000,\n%s0.000000,0.000000,1.000000,0.000000,\n%s0.000000,0.000000,0.000000,1.000000;;\n%s}\n" % (XSI.Indent * 1, XSI.Indent * 2, XSI.Indent * 2, XSI.Indent * 2, XSI.Indent * 2, XSI.Indent * 1))
			XSI.IndentLvl += 1
		
		# Write frames.
		for Frm in InitObjects:
			AddFrame(XSI, Frm)
		
		# If we added a dummyroot, close it off.
		if XSI.AddDummyRoot == 1:
			XSI.File.write("}\n")
			XSI.IndentLvl -= 1
		
		# Animations
		if XSI.ExportAnimations:
			AnimatedObjects = [FrmObj for FrmObj in AllObjects if FrmObj.hasAnim]
			if AnimatedObjects:
				print("\nAnimationSet")
				XSI.File.write("\nAnimationSet {\n")
				XSI.IndentLvl += 1
				for AnimObj in AnimatedObjects:
					name = LegalName(AnimObj.name)
					print("%sAnimation anim-%s\t(P=%d, R=%d)" % ((" "*XSI.IndentLvl), name, AnimObj.Anim.PosKeyFrameCount, AnimObj.Anim.RotKeyFrameCount))
					XSI.File.write("%sAnimation anim-%s {\n" % ((XSI.Indent*XSI.IndentLvl), name))
					XSI.IndentLvl += 1
					XSI.File.write("%s{frm-%s}\n" % ((XSI.Indent*XSI.IndentLvl), name))
					AddAnimation(XSI, AnimObj.Anim)
					XSI.IndentLvl -= 1
					XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))
				XSI.IndentLvl -= 1
				XSI.File.write("}\n")
		
		# Envelopes
		if XSI.ExportEnvelopes:
			EnvelopedObjects = [FrmObj for FrmObj in AllObjects if FrmObj.envelopes]
			if EnvelopedObjects:
				print("\nSI_EnvelopeList")
				XSI.File.write("\nSI_EnvelopeList {\n")
				XSI.IndentLvl += 1
				envCount = 0
				for EnvObj in EnvelopedObjects:
					for Env in EnvObj.envelopes:
						envCount += 1
				XSI.File.write("%s%d;\n" % ((XSI.Indent*XSI.IndentLvl), envCount))
				for EnvObj in EnvelopedObjects:
					name = LegalName(EnvObj.name)
					for Env in EnvObj.envelopes:
						boneName = LegalName(Env)
						print("%sSI_Envelope\t(frm-%s:frm-%s (%d))" % ((" "*XSI.IndentLvl), name, boneName, len(EnvObj.envelopes)))
						XSI.File.write("%sSI_Envelope {\n" % ((XSI.Indent*XSI.IndentLvl)))
						XSI.IndentLvl += 1
						XSI.File.write("%s\"frm-%s\";\n" % ((XSI.Indent*XSI.IndentLvl), name))
						XSI.File.write("%s\"frm-%s\";\n" % ((XSI.Indent*XSI.IndentLvl), boneName))
						vertCount = len(EnvObj.envelopes[Env][0])
						XSI.File.write("%s%d;\n" % ((XSI.Indent*XSI.IndentLvl), vertCount))
						for Index in range(vertCount):
							XSI.File.write("%s%d;%f;" % ((XSI.Indent*XSI.IndentLvl), EnvObj.envelopes[Env][0][Index], EnvObj.envelopes[Env][1][Index]))
							if Index == vertCount-1:
								XSI.File.write(";\n")
							else:
								XSI.File.write(",\n")
						XSI.IndentLvl -= 1
						XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))
				XSI.IndentLvl -= 1
				XSI.File.write("}\n")
		
		# End
		print("BZ2XSI exported.")
		
		XSI.File.close()
	else:
		print("<FRM CLASS DEBUG:>")
		for FrmObj in InitObjects:
			PrintChildrenDebug(FrmObj)
		print("<END>")

def PrintChildrenDebug(FrmObj, IndentLvl = 0, UseLegalName = True):
	if UseLegalName:
		name = LegalName(FrmObj.name)
	else:
		name = FrmObj.name
	
	Basic = ("> %s\t%s%s" % (FrmObj.VType[0:5], (" "*IndentLvl), name))
	Extra = ""
	Children = ("   Chldrn(%d: %s)" % (len(FrmObj.children), str([Child.name for Child in FrmObj.children]).strip('[]')))
	Animation = ("   KeyFrms(P%d, R%d)" % (len(FrmObj.Anim.Pos), len(FrmObj.Anim.Rot)))
	
	# ---
	#~ Extra += Children
	Extra += Animation
	# ---
	
	print(Basic + Extra)
	
	# Temporary extra information
	#~ print(FrmObj.Anim.PosKeyFrames)
	#~ print(FrmObj.Anim.RotKeyFrames)
	# ---
	
	for SubFrmObj in FrmObj.children:
		PrintChildrenDebug(SubFrmObj, IndentLvl+1)

def CreateObjects():
	AllObjects = []
	InitObjects = []
	for Obj in bpy.data.objects:
		if Obj.parent is None and (Obj.type == "MESH" or Obj.type == "ARMATURE" or Obj.type == "EMPTY"):
			InitObjects += [Frm(Obj.name, Obj.type, Obj, None)]
	
	for FrmObj in InitObjects:
		AllObjects += [FrmObj] + GetChildrenRecurse(FrmObj)
	
	return AllObjects

def GetChildrenRecurse(FrmObj):
	Objects = []
	for Child in FrmObj.children:
		Objects += [Child] + GetChildrenRecurse(Child)
	return Objects

def LegalName(Str):
	return re.sub("[\s\.]", "_", Str.lower())

def GetDataBoneMatrix(DataBone):
	BoneMatrix = Matrix()
	Par = DataBone.parent
	if Par:
		BoneMatrix *= Matrix(Par.matrix_local).inverted()
		BoneMatrix *= Matrix(DataBone.matrix_local)
	else:
		BoneMatrix *= Matrix(DataBone.matrix_local)
	return BoneMatrix

def GetPoseBoneMatrix(PoseBone):
	BoneMatrix = Matrix()
	Par = PoseBone.parent
	if Par:
		BoneMatrix = Matrix(Par.matrix).inverted()
		BoneMatrix *= Matrix(PoseBone.matrix)
	else:
		BoneMatrix = Matrix(PoseBone.matrix)
	return BoneMatrix

def AddFrame(XSI, FrmObj):
	name = LegalName(FrmObj.name)
	
	# Frame Start
	XSI.File.write("%sFrame frm-%s {\n" % ((XSI.Indent*XSI.IndentLvl), name))
	XSI.IndentLvl += 1
	
	if not FrmObj.parent:
		AddMatrix(XSI, FrmObj.matrix, False, axisConv)
	else:
		AddMatrix(XSI, FrmObj.matrix, False)
	
	if FrmObj.poseMatrix:
		AddMatrix(XSI, FrmObj.poseMatrix, True)
	if FrmObj.mesh:
		AddMesh(XSI, FrmObj, name)
	elif XSI.BoneMesh and FrmObj.VType == "BONE":
		AddBoneMesh(XSI, FrmObj.realObject[2])
	
	# Write Children
	XSI.IndentLvl += 1
	for SubFrmObj in FrmObj.children:
		AddFrame(XSI, SubFrmObj)
	XSI.IndentLvl -= 1
	
	# Frame End
	XSI.IndentLvl -= 1
	XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))

def AddAnimation(XSI, Anim):
	if Anim.PosKeyFrameCount:
		# XSI Anim Type: 2, Translational.
		XSI.File.write("%sSI_AnimationKey {\n" % (XSI.Indent*XSI.IndentLvl))
		XSI.IndentLvl += 1
		XSI.File.write("%s2;\n" % (XSI.Indent*XSI.IndentLvl))
		XSI.File.write("%s%d;\n" % ((XSI.Indent*XSI.IndentLvl), Anim.PosKeyFrameCount))
		for Index in range(Anim.PosKeyFrameCount-1):
				XSI.File.write("%s%d;3;%f,%f,%f;;,\n" % ((XSI.Indent*XSI.IndentLvl), Anim.PosKeyFrames[Index], Anim.Pos[Index][0], Anim.Pos[Index][1], Anim.Pos[Index][2]))
		else:
			Index += 1
			XSI.File.write("%s%d;3;%f,%f,%f;;;\n" % ((XSI.Indent*XSI.IndentLvl), Anim.PosKeyFrames[Index], Anim.Pos[Index][0], Anim.Pos[Index][1], Anim.Pos[Index][2]))
		XSI.IndentLvl -= 1
		XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))
	if Anim.RotKeyFrameCount:
		# XSI Anim Type: 0, Quaternion.
		XSI.File.write("%sSI_AnimationKey {\n" % (XSI.Indent*XSI.IndentLvl))
		XSI.IndentLvl += 1
		XSI.File.write("%s0;\n" % (XSI.Indent*XSI.IndentLvl))
		XSI.File.write("%s%d;\n" % ((XSI.Indent*XSI.IndentLvl), Anim.RotKeyFrameCount))
		for Index in range(Anim.RotKeyFrameCount-1):
			XSI.File.write("%s%d;4;%f,%f,%f,%f;;,\n" % ((XSI.Indent*XSI.IndentLvl), Anim.RotKeyFrames[Index], Anim.Rot[Index][0], Anim.Rot[Index][1], Anim.Rot[Index][2], Anim.Rot[Index][3]))
		else:
			Index += 1
			XSI.File.write("%s%d;4;%f,%f,%f,%f;;;\n" % ((XSI.Indent*XSI.IndentLvl), Anim.RotKeyFrames[Index], Anim.Rot[Index][0], Anim.Rot[Index][1], Anim.Rot[Index][2], Anim.Rot[Index][3]))
		XSI.IndentLvl -= 1
		XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))

def AddMesh(XSI, FrmObj, name):
	Mesh = FrmObj.mesh
	
	# Mesh Begin
	XSI.File.write("%sMesh %s {\n" % ((XSI.Indent*XSI.IndentLvl), name))
	XSI.IndentLvl += 1
	
	# Mesh Vertex List
	AddMeshVerts(XSI, Mesh)
	
	# Mesh Polygons
	TerminateChar = ","
	PolyCount = len(Mesh.polygons)
	XSI.File.write("%s%d;\n" % ((XSI.Indent*XSI.IndentLvl), PolyCount))
	for PolyIndex, Polygon in enumerate(Mesh.polygons):
		if PolyIndex >= PolyCount-1:
			TermwnateChar = ";"
		VertexCount = len(Polygon.vertices)
		XSI.File.write("%s%d;" % ((XSI.Indent*XSI.IndentLvl), VertexCount))
		for i in range(VertexCount):
			XSI.File.write("%d" % Polygon.vertices[i])
			if i >= VertexCount-1:
				XSI.File.write(";")
			else:
				XSI.File.write(",")
		XSI.File.write("%s\n" % TerminateChar)
	XSI.File.write("\n")
	
	# The rest of the shit, in order occirding to stock XSI apgun00.xsi.
	AddMaterialList(XSI, Mesh)
	AddMeshNormals(XSI, Mesh)
	if XSI.ExportVertColors and Mesh.vertex_colors.active:
		if FrmObj.envelopes:
			print("!%s%s: Vert colors do not work on enveloped meshes in BZ2! Exporting anyways..." % ((" "*(XSI.IndentLvl-1)), LegalName(FrmObj.name)))
		AddVertexColors(XSI, Mesh)
	if XSI.ExportUV and Mesh.uv_layers.active:
		AddTextureCoords(XSI, Mesh)
	
	# Mesh End
	XSI.IndentLvl -= 1
	XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))

def AddMeshVerts(XSI, Mesh):
	TerminateChar = ","
	VertexCount = len(Mesh.vertices)
	XSI.File.write("%s%d;\n" % ((XSI.Indent*XSI.IndentLvl), VertexCount))
	for i in range(VertexCount):
		Vert = Mesh.vertices[i].co.xyz
		if i >= VertexCount-1:
			TerminateChar = ";\n"
		XSI.File.write("%s%f;%f;%f;%s\n" % ((XSI.Indent*XSI.IndentLvl), Vert[0], Vert[1], Vert[2], TerminateChar))

def AddMeshNormals(XSI, Mesh):
	# SI_MeshNormals Begin
	XSI.File.write("%sSI_MeshNormals {\n" % (XSI.Indent*XSI.IndentLvl))
	XSI.IndentLvl += 1
	VertexCount = GetMeshVertexCount(Mesh)
	
	# Vertex List
	TerminateChar = ","
	XSI.File.write("%s%d;\n" % ((XSI.Indent*XSI.IndentLvl), VertexCount))
	i = 0
	for Polygon in Mesh.polygons:
		Vertices = list(Polygon.vertices)
		
		for Vertex in [Mesh.vertices[Vertex] for Vertex in Vertices]:
			if Polygon.use_smooth:
				Normal = Vertex.normal
			else:
				Normal = Polygon.normal
			if i >= VertexCount-1:
				TerminateChar = ";\n"
			i += 1
			XSI.File.write("%s%f;%f;%f;%s\n" % ((XSI.Indent*XSI.IndentLvl), Normal[0], Normal[1], Normal[2], TerminateChar))
	
	# Polygons
	TerminateChar = ","
	XSI.File.write("%s%d;\n" % ((XSI.Indent*XSI.IndentLvl), len(Mesh.polygons)))
	i = 0
	PolyIndex = 0
	for Polygon in Mesh.polygons:
		XSI.File.write("%s%d;%d;" % ((XSI.Indent*XSI.IndentLvl), PolyIndex, len(Polygon.vertices)))
		PolyIndex += 1
		for Vertex in Polygon.vertices:
			XSI.File.write("%d;" % i)
			if i >= VertexCount-1:
				TerminateChar = ";"
			i += 1
		XSI.File.write("%s\n" % TerminateChar)
	
	# SI_MeshNormals End
	XSI.IndentLvl -= 1
	XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))

def AddTextureCoords(XSI, Mesh):
	# SI_MeshTextureCoords Begin
	XSI.File.write("%sSI_MeshTextureCoords {\n" % (XSI.Indent*XSI.IndentLvl))
	XSI.IndentLvl += 1
	VertexCount = GetMeshVertexCount(Mesh)
	
	# Vertex List
	TerminateChar = ","
	XSI.File.write("%s%d;\n" % ((XSI.Indent*XSI.IndentLvl), VertexCount))
	i = 0
	for UV in Mesh.uv_layers.active.data:
		if i >= VertexCount-1:
			TerminateChar = ";\n"
		i += 1
		XSI.File.write("%s%f;%f;%s\n" % ((XSI.Indent*XSI.IndentLvl), UV.uv[0], UV.uv[1], TerminateChar))
	
	# Polygons (Same as MeshNormals.)
	TerminateChar = ","
	XSI.File.write("%s%d;\n" % ((XSI.Indent*XSI.IndentLvl), len(Mesh.polygons)))
	i = 0
	PolyIndex = 0
	for Polygon in Mesh.polygons:
		XSI.File.write("%s%d;%d;" % ((XSI.Indent*XSI.IndentLvl), PolyIndex, len(Polygon.vertices)))
		PolyIndex += 1
		for Vertex in Polygon.vertices:
			XSI.File.write("%d;" % i)
			if i >= VertexCount-1:
				TerminateChar = ";"
			i += 1
		XSI.File.write("%s\n" % TerminateChar)
	
	# SI_MeshTextureCoords End
	XSI.IndentLvl -= 1
	XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))

def AddVertexColors(XSI, Mesh):
	# SI_MeshVertexColors Begin
	XSI.File.write("%sSI_MeshVertexColors {\n" % (XSI.Indent*XSI.IndentLvl))
	XSI.IndentLvl += 1
	VertexCount = GetMeshVertexCount(Mesh)
	
	# Vertex List
	XSI.File.write("%s%d;\n" % ((XSI.Indent*XSI.IndentLvl), VertexCount))
	VertIndex = 0
	for Polygon in Mesh.polygons:
		for Vert in range(len(Polygon.vertices)-1):
			XSI.File.write("%s%f;%f;%f;%f;,\n" % ((XSI.Indent*XSI.IndentLvl), Mesh.vertex_colors.active.data[VertIndex].color[0], Mesh.vertex_colors.active.data[VertIndex].color[1], Mesh.vertex_colors.active.data[VertIndex].color[2], 1))
			VertIndex += 1
		else:
			XSI.File.write("%s%f;%f;%f;%f;;\n" % ((XSI.Indent*XSI.IndentLvl), Mesh.vertex_colors.active.data[VertIndex].color[0], Mesh.vertex_colors.active.data[VertIndex].color[1], Mesh.vertex_colors.active.data[VertIndex].color[2], 1))
			VertIndex += 1
	XSI.File.write("\n")
	
	# Polygons (Same as MeshNormals.)
	TerminateChar = ","
	XSI.File.write("%s%d;\n" % ((XSI.Indent*XSI.IndentLvl), len(Mesh.polygons)))
	i = 0
	PolyIndex = 0
	for Polygon in Mesh.polygons:
		XSI.File.write("%s%d;%d;" % ((XSI.Indent*XSI.IndentLvl), PolyIndex, len(Polygon.vertices)))
		PolyIndex += 1
		for Vertex in Polygon.vertices:
			XSI.File.write("%d;" % i)
			if i >= VertexCount-1:
				TerminateChar = ";"
			i += 1
		XSI.File.write("%s\n" % TerminateChar)
	
	# SI_MeshVertexColors Begin
	XSI.IndentLvl -= 1
	XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))

def AddMaterialList(XSI, Mesh):
	# MeshMaterialList Begin
	XSI.File.write("%sMeshMaterialList {\n" % (XSI.Indent*XSI.IndentLvl))
	XSI.IndentLvl += 1
	
	# Copied from io_export_directx_x.py, as it's identical as far as I know.
	# ---
	Materials = Mesh.materials
	if Materials.keys():
		MaterialIndexes = {}
		for Polygon in Mesh.polygons:
			if Materials[Polygon.material_index] not in MaterialIndexes:
				MaterialIndexes[Materials[Polygon.material_index]] = len(MaterialIndexes)
		
		PolygonCount = len(Mesh.polygons)
		Index = 0
		XSI.File.write("%s%d;\n%s%d;\n" % ((XSI.Indent*XSI.IndentLvl), len(MaterialIndexes), (XSI.Indent*XSI.IndentLvl), PolygonCount))
		for Polygon in Mesh.polygons:
			XSI.File.write("%s%d" % ((XSI.Indent*XSI.IndentLvl), MaterialIndexes[Materials[Polygon.material_index]]))
			Index += 1
			if Index == PolygonCount:
				XSI.File.write(";;\n")
			else:
				XSI.File.write(",\n")
		
		Materials = [Item[::-1] for Item in MaterialIndexes.items()]
		Materials.sort()
		for Material in Materials:
			AddMaterial(XSI, Material[1])
	else:
		XSI.File.write("%s1;\n%s1;\n%s0;;\n" % ((XSI.Indent*XSI.IndentLvl), (XSI.Indent*XSI.IndentLvl), (XSI.Indent*XSI.IndentLvl)))
		AddMaterial(XSI, None)
	# ---
	
	# MeshMaterialList End
	XSI.IndentLvl -= 1
	XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))

def AddMaterial(XSI, Material):
	# SI_Material Begin
	XSI.File.write("%sSI_Material {\n" % (XSI.Indent*XSI.IndentLvl))
	XSI.IndentLvl += 1
	
	# Copied from io_export_directx_x.py, as it's identical as far as I know.
	# ---
	if Material:
		Diffuse = list(Vector(Material.diffuse_color) * Material.diffuse_intensity)
		Diffuse.append(Material.alpha)
		Specularity = 1000 * (Material.specular_hardness - 1.0) / (511.0 - 1.0) # Map Blender's range of 1 - 511 to 0 - 1000
		Specular = list(Vector(Material.specular_color) * Material.specular_intensity)
		
		XSI.File.write("%s%f;%f;%f;%f;;\n" % ((XSI.Indent*XSI.IndentLvl), Diffuse[0], Diffuse[1], Diffuse[2], Diffuse[3]))
		XSI.File.write("%s%f;\n" % ((XSI.Indent*XSI.IndentLvl), Specularity))
		XSI.File.write("%s%f;%f;%f;;\n" % ((XSI.Indent*XSI.IndentLvl), Specular[0], Specular[1], Specular[2]))
	else:
		XSI.File.write("%s0.800000;0.800000;0.800000;0.800000;;\n" % (XSI.Indent*XSI.IndentLvl))
		XSI.File.write("%s96.078431;\n" % (XSI.Indent*XSI.IndentLvl)) # 1000 * (50 - 1) / (511 - 1)
		XSI.File.write("%s0.500000;0.500000;0.500000;;\n" % (XSI.Indent*XSI.IndentLvl))
	XSI.File.write("%s0.000000;0.000000;0.000000;;\n" % (XSI.Indent*XSI.IndentLvl))
	XSI.File.write("%s5;\n" % (XSI.Indent*XSI.IndentLvl))
	XSI.File.write("%s0.500000;0.500000;0.500000;;\n" % (XSI.Indent*XSI.IndentLvl))
	TextureFileName = GetMaterialTextureFileName(Material)
	if TextureFileName:
		XSI.File.write("%sSI_Texture2D {\n%s\"%s\";\n%s}\n" % ((XSI.Indent*XSI.IndentLvl), (XSI.Indent*(XSI.IndentLvl+1)), TextureFileName, (XSI.Indent*XSI.IndentLvl)))
	# ---
	
	# SI_Material End
	XSI.IndentLvl -= 1
	XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))

def AddMatrix(XSI, Mat, Pose = False, axisConv = None):
	if Pose:
		XSI.File.write("%sSI_FrameBasePoseMatrix {\n" % (XSI.Indent*XSI.IndentLvl))
	else:
		XSI.File.write("%sFrameTransformMatrix {\n" % (XSI.Indent*XSI.IndentLvl))
	XSI.IndentLvl += 1
	if axisConv:
		NMat = Mat*axisConv
		NMat[1][3] = Mat[2][3]
		NMat[2][3] = Mat[1][3]
	else:
		NMat = Mat
	XSI.File.write("%s%f,%f,%f,%f,\n" % ((XSI.Indent*XSI.IndentLvl), NMat[0][0], NMat[1][0], NMat[2][0], NMat[3][0]))
	XSI.File.write("%s%f,%f,%f,%f,\n" % ((XSI.Indent*XSI.IndentLvl), NMat[0][1], NMat[1][1], NMat[2][1], NMat[3][1]))
	XSI.File.write("%s%f,%f,%f,%f,\n" % ((XSI.Indent*XSI.IndentLvl), NMat[0][2], NMat[1][2], NMat[2][2], NMat[3][2]))
	XSI.File.write("%s%f,%f,%f,%f;;\n%s}\n" % ((XSI.Indent*XSI.IndentLvl), NMat[0][3], NMat[1][3], NMat[2][3], NMat[3][3], (XSI.Indent*(XSI.IndentLvl-1))))
	XSI.IndentLvl -= 1

# From io_export_directx_x.py.
def GetMaterialTextureFileName(Material):
	if Material:
		# Create a list of Textures that have type "IMAGE"
		ImageTextures = [Material.texture_slots[TextureSlot].texture for TextureSlot in Material.texture_slots.keys() if Material.texture_slots[TextureSlot].texture.type == "IMAGE"]
		# Refine a new list with only image textures that have a file source
		ImageFiles = [bpy.path.basename(Texture.image.filepath) for Texture in ImageTextures if getattr(Texture.image, "source", "") == "FILE"]
		if ImageFiles:
			return ImageFiles[0]
	return None

# From io_export_directx_x.py
def GetMeshVertexCount(Mesh):
	VertexCount = 0
	for Polygon in Mesh.polygons:
		VertexCount += len(Polygon.vertices)
	return VertexCount

# Made this function to help debug bones more visually.
# Also exports bone group color as material diffuse color.
def AddBoneMesh(XSI, Bone):
	Radius = Bone.length*0.1
	Base = Bone.length*0.2
	Tip = Bone.length
	if Bone.bone_group:
		Color = Bone.bone_group.colors.active
	else:
		Color = (1.000000, 1.000000, 1.000000)
	Alpha = 1.000000
	
	XSI.File.write("%sMesh %s {\n" % ((XSI.Indent*XSI.IndentLvl), LegalName(Bone.name)))
	XSI.IndentLvl += 1 # +Mesh
	XSI.File.write("%s6;\n" % ((XSI.Indent*XSI.IndentLvl)))
	XSI.File.write("%s-%f;%f;-%f;,\n" % ((XSI.Indent*XSI.IndentLvl), Radius, Base, Radius))
	XSI.File.write("%s0.000000;0.000000;0.000000;,\n" % (XSI.Indent*XSI.IndentLvl))
	XSI.File.write("%s%f;%f;-%f;,\n" % ((XSI.Indent*XSI.IndentLvl), Radius, Base, Radius))
	XSI.File.write("%s-%f;%f;%f;,\n" % ((XSI.Indent*XSI.IndentLvl), Radius, Base, Radius))
	XSI.File.write("%s%f;%f;%f;,\n" % ((XSI.Indent*XSI.IndentLvl),  Radius, Base, Radius))
	XSI.File.write("%s0.000000;%f;0.000000;,\n" % ((XSI.Indent*XSI.IndentLvl), Tip))
	XSI.File.write("!8;!3;1,4,2;,!3;0,3,1;,!3;3,4,1;,!3;0,1,2;,!3;4,3,5;,!3;0,2,5;,!3;3,0,5;,!3;2,4,5;,".replace("!", "\n%s" % (XSI.Indent*XSI.IndentLvl)))
	XSI.File.write("\n\n%sMeshMaterialList {\n%s" % ((XSI.Indent*XSI.IndentLvl), (XSI.Indent*(XSI.IndentLvl+1))))
	XSI.IndentLvl += 1 # +MeshMaterialList
	XSI.File.write("1;!8;!0,!0,!0,!0,!0,!0,!0,!0;;!SI_Material {".replace("!", "\n%s" % (XSI.Indent*XSI.IndentLvl)))
	XSI.IndentLvl += 1 # +SI_Material
	XSI.File.write("!%f;%f;%f;%f;;!96.078431;!0.500000;0.500000;0.500000;;!%f;%f;%f;;!5;!0.500000;0.500000;0.500000;;\n".replace("!", "\n%s" % (XSI.Indent*XSI.IndentLvl)) % (Color[0], Color[1], Color[2], Alpha, Color[0]/2, Color[1]/2, Color[2]/2))
	XSI.IndentLvl -= 1 # -SI_Material
	XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))
	XSI.IndentLvl -= 1 # -MeshMaterialList
	XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))
	XSI.IndentLvl += 1 # +SI_MeshNormals
	XSI.File.write("%sSI_MeshNormals {" % (XSI.Indent*XSI.IndentLvl))
	XSI.File.write("!24;!0.707107;-0.000000;-0.707107;,!0.707107;-0.000000;-0.707107;,!0.707107;-0.000000;-0.707107;,!-0.707107;-0.000000;-0.707107;,!-0.707107;-0.000000;-0.707107;,!-0.707107;-0.000000;-0.707107;,!0.000000;0.707107;-0.707107;,!0.000000;0.707107;-0.707107;,!0.000000;0.707107;-0.707107;,!0.000000;-0.707107;-0.707107;,!0.000000;-0.707107;-0.707107;,!0.000000;-0.707107;-0.707107;,!-0.000000;0.989949;0.141421;,!-0.000000;0.989949;0.141421;,!-0.000000;0.989949;0.141421;,!-0.000000;-0.989949;0.141421;,!-0.000000;-0.989949;0.141421;,!-0.000000;-0.989949;0.141421;,!-0.989949;-0.000000;0.141421;,!-0.989949;-0.000000;0.141421;,!-0.989949;-0.000000;0.141421;,!0.989949;-0.000000;0.141421;,!0.989949;-0.000000;0.141421;,!0.989949;-0.000000;0.141421;;\n".replace("!", "\n%s" % (XSI.Indent*XSI.IndentLvl)))
	XSI.File.write("!8;!0;3;0;1;2;,!1;3;3;4;5;,!2;3;6;7;8;,!3;3;9;10;11;,!4;3;12;13;14;,!5;3;15;16;17;,!6;3;18;19;20;,!7;3;21;22;23;;\n".replace("!", "\n%s" % (XSI.Indent*XSI.IndentLvl)))
	XSI.IndentLvl -= 1 # -SI_MeshNormals
	XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))
	XSI.IndentLvl -= 1 # - Mesh
	XSI.File.write("%s}\n" % (XSI.Indent*XSI.IndentLvl))

class BZ2XSI_Export(bpy.types.Operator):
	bl_idname = "export_scene.bz2xsi"
	bl_label = "Export BZ2XSI"
	
	filepath = StringProperty(subtype="FILE_PATH")
	
	AddDummyRoot = BoolProperty(
	name="Add Dummyroot",
	description="Add an hp_dummyroot frame to parent all frames.",
	default=False)
	
	IndentTypes = (
	("N", "None", ""),("\t", "1 Tab", ""),(" " * 1, "1 Space", ""),
	(" " * 2, "2 Spaces", ""),(" " * 3, "3 Spaces", ""),(" " * 4, "4 Spaces", ""),
	(" " * 5, "5 Spaces", ""),(" " * 6, "6 Spaces", ""),(" " * 8, "8 Spaces", ""),
	)
	Indent = EnumProperty(
	name="Indentation",
	description="",
	items=IndentTypes,
	default="\t")
	
	BoneMesh = BoolProperty(
	name="Create Bone Mesh",
	description="Generates a mesh for armature bones. Added for debugging purposes.",
	default=False)
	
	ExportUV = BoolProperty(
	name="Export UV's",
	description="Export UV texture coordinates.",
	default=True)
	
	ExportVertColors = BoolProperty(
	name="Export Vert Colors",
	description="Export vertex colors.",
	default=True)
	
	ExportAnimations = BoolProperty(
	name="Export Animations",
	description="Export animations.",
	default=True)
	
	ExportEnvelopes = BoolProperty(
	name="Export Envelopes",
	description="Export bone envelopes.",
	default=True)
	
	def execute(self, context):
		FilePath = bpy.path.ensure_ext(self.filepath, ".xsi")
		
		XSI = BZ2XSI_Settings(context, FilePath, AddDummyRoot=self.AddDummyRoot, Indent=self.Indent, BoneMesh=self.BoneMesh, ExportUV=self.ExportUV, ExportVertColors=self.ExportVertColors, ExportAnimations=self.ExportAnimations, ExportEnvelopes=self.ExportEnvelopes)
		
		# Had to use "N" due to "" making the menu item disabled.
		if XSI.Indent == "N":
			XSI.Indent = ""
		
		ExportXSI(XSI)
		
		return {"FINISHED"}
	def invoke(self, context, event):
		if not self.filepath:
			self.filepath = bpy.path.ensure_ext(bpy.data.filepath, ".xsi")
		WindowManager = context.window_manager
		WindowManager.fileselect_add(self)
		return {"RUNNING_MODAL"}

def addMenuItem(self, context):
	self.layout.operator(BZ2XSI_Export.bl_idname, text="BZ2XSI (.xsi)")

def register():
	bpy.utils.register_class(BZ2XSI_Export)
	bpy.types.INFO_MT_file_export.append(addMenuItem)

def unregister():
	bpy.utils.unregister_class(BZ2XSI_Export)
	bpy.types.INFO_MT_file_export.remove(addMenuItem)
